/* 
 * new.C - A memory management system.
 * 
 * Author:	Robert Mecklenburg
 * 		Computer Science Dept.
 * 		University of Utah
 * Date:	Wed May 26 1993
 *
 * Copyright (c) 1993,1994 The University of Utah and
 * the Center for Software Science (CSS).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the Center
 * for Software Science at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSS ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSS DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSS requests users of this software to return to css-dist@cs.utah.edu any
 * improvements that they make and grant CSS redistribution rights.
 */

#ifdef GC
#include <gc.h>
#endif

#include <limits.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <iostream.h>
#include <new.H>


/*****************************************************************
 * mem_list
 */
mem_list_c	mem_list;


/*****************************************************************
 * mem_list_c
 */
int		mem_list_c::count 	= 0;
int		mem_list_c::watch_new	= 0;
int		mem_list_c::watch_del	= 0;
int		mem_list_c::max_size	= 0;


/*****************************************************************
 * operator new
 *
 * The global C++ memory allocation function.  Here we allocate the
 * amount of memory requested by the application, plus enough room for
 * the header and cookie.  We then insert the new memory into the list
 * and bump the pointer, returning application useable memory.
 */
void *
operator new( size_t s )
{
#ifdef GC
    mem_elem_c * ret = (mem_elem_c *)GC_MALLOC( s +
						sizeof( mem_elem_c ) +
						sizeof( char ) );
#else
    mem_elem_c * ret = (mem_elem_c *)malloc( s +
					     sizeof( mem_elem_c ) +
					     sizeof( char ) );
#endif
    assert( ret );
    mem_list.insert( ret, s, NULL, 0 );
    ret++;
    return ret;
}


/*****************************************************************
 * operator new
 *
 * The global C++ memory allocation function.  Here we allocate the
 * amount of memory requested by the application, plus enough room for
 * the header and cookie.  We then insert the new memory into the list
 * and bump the pointer, returning application useable memory.
 */
void *
operator new( size_t s, const char * file, int line )
{
#ifdef GC
# if DEBUG
    mem_elem_c * ret = (mem_elem_c *)GC_debug_malloc( s +
						      sizeof( mem_elem_c ) +
						      sizeof( char ),
						      (char *)file, line );
# else
    mem_elem_c * ret = (mem_elem_c *)GC_MALLOC( s +
						sizeof( mem_elem_c ) +
						sizeof( char ) );
# endif
#else
    mem_elem_c * ret = (mem_elem_c *)malloc( s +
					     sizeof( mem_elem_c ) +
					     sizeof( char ) );
#endif
    assert( ret );
    mem_list.insert( ret, s, file, line );
    ret++;
    return ret;
}


#if defined (__GNUG__) || defined(_HAS_NEW_ARRAY_OPERATOR_)
/*****************************************************************
 * operator new[]
 *
 * The global C++ array memory allocation function.  Here we allocate
 * the amount of memory requested by the application, plus enough room
 * for the header and cookie.  We then insert the new memory into the
 * list and bump the pointer, returning application useable memory.
 */
void *
operator new[]( size_t size, const char * file, int line )
{
#ifdef GC
# if DEBUG
    mem_elem_c * ret = (mem_elem_c *)GC_debug_malloc( s +
						      sizeof( mem_elem_c ) +
						      sizeof( char ),
						      (char *)file, line );
# else
    mem_elem_c * ret = (mem_elem_c *)GC_MALLOC( s +
						sizeof( mem_elem_c ) +
						sizeof( char ) );
# endif
#else
    mem_elem_c * ret = (mem_elem_c *)malloc( size +
					     sizeof( mem_elem_c ) +
					     sizeof( char ) );
#endif
    assert( ret );
    mem_list.insert( ret, size, file, line );
    ret++;
    return ret;
}
#endif


/*****************************************************************
 * operator delete
 *
 * The global C++ memory deallocation function.  If the pointer is
 * non-null we find the start of the memory unit and remove it from
 * the linked list, then free it.
 */
void
operator delete( void * v )
{
    if ( v )
    {
	mem_elem_c * ret = ((mem_elem_c *)v) - 1;
	mem_list.remove( ret );
#ifdef GC
	GC_FREE( ret );
#else
	free( ret );
#endif
    }
}


/*****************************************************************
 * mem_elem_c::set_free
 *
 * Set or clear the freed bit in a memory element.
 */
void
mem_elem_c::set_free( int f )
{
    if ( f )
	size_ |= FREE_MASK;
    else
	size_ &= ~FREE_MASK;
}


/*****************************************************************
 * mem_elem_c::set_control
 *
 * Set the control bits in a memory element.  These consist of the
 * unique identifier, the size, and freed bit.
 */
void
mem_elem_c::set_control( int cnt, int sz, int fr, const char * fi, int li )
{
    /* Set counter. */
    assert( cnt > 0 && cnt < INT_MAX );
    count_ = cnt;

    /* Set the size. */
    if ( sz > mem_list_c::max_size )
	mem_list_c::max_size = sz;

    if ( sz <= SIZE_MASK )
	size_ = (size_ & ~SIZE_MASK) | sz;
    else
    {
	fprintf( stderr, "Object size %d too large for size field.\n", sz );
	abort();
    }

    set_free( fr );
    file_ = fi;
    line_ = li;
}


/*****************************************************************
 * mem_elem_c::clear
 *
 * Set the memory associated with a memory element to the indicated
 * value.
 */
void
mem_elem_c::clear( unsigned char val )
{
    if ( size() )
	memset( this + 1, val, size() );
}


/*****************************************************************
 * mem_elem_c::valid
 *
 * Check that the memory element is a legal one.
 */
int
mem_elem_c::valid() const
{
    const char * msg = NULL;

    if ( this == NULL )
	msg = "Delete of this == NULL.";

    /* Count is reasonable. */
    else if ( count() > mem_list_c::count )
	msg = "Bogus count.";

    /* Size is reasonable. */
    else if ( size() > mem_list_c::max_size )
	msg = "Bogus size.";

#ifndef GC
    /* Prev address is illegal if it is the head of the list and prev
     * is non-null or it is not the first and prev is null.
     */
    else if ( mem_list.head() == this && prev_ != NULL ||
	      mem_list.head() != this && prev_ == NULL )
	msg = "Invalid prev_ pointer.";

    /* The prev_ address is properly aligned. */
    else if ( ((unsigned long)prev_ & 0x7) != 0 )
	msg = "Misaligned prev_ pointer.";

    /* This element is in the proper order. */
    else if ( prev_ && prev_->count() <= count() )
	msg = "Bad list order.";

    /* The next_ address is properly aligned. */
    else if ( ((unsigned long)next_ & 0x7) != 0 )
	msg = "Misaligned next_ pointer.";

    /* This element is in the proper order. */
    else if ( next_ && next_->count() >= count() )
	msg = "Bad list order.";
#endif

    /* Check for memory overruns. */
    else if ( check_cookie() == 0 )
	msg = "Cookie damage.";

    /* Check not already freed. */
    else if ( freed() )
	msg = "Already freed.";

    if ( msg )
    {
	fprintf( stderr, "%s(%d): %s\n", __FILE__, __LINE__, msg );
	report( 0, 0 );
	return 0;
    }
    else
	return 1;
}


/*****************************************************************
 * mem_elem_c::set_cookie
 *
 * Set the cookie at the end of user memory.  The cookie is set to a
 * changing, but identifiable value -- the lower 8 bits of the count_
 * field.
 */
void
mem_elem_c::set_cookie()
{
    unsigned char * cp = (unsigned char *)(this + 1) + size();
    *cp = count() & 0xFF;
}


/*****************************************************************
 * mem_elem_c::check_cookie
 *
 * Check that the cookie is unmolested.  If the cookie does not match
 * the lower 8 bits of the count_ field, then the application has
 * overrun the bounds of allocated memory.  Return true if the cookie
 * is ok.
 */
int
mem_elem_c::check_cookie() const
{
    unsigned char * cp = (unsigned char *)(this + 1) + size();
    return *cp == (count() & 0xFF);
}


/*****************************************************************
 * mem_elem_c::report
 *
 */
void
mem_elem_c::report( int omit_hdr, int chk ) const
{
    if ( this == NULL )
	return;

    if ( chk )
	assert( valid() );

    if ( ! omit_hdr )
	fprintf( stderr, " seq_num    size    address  line file\n" );

    if ( file_ )
	fprintf( stderr, "%8d%8d 0x%.8x %5d %s\n",
		 count(), size(), this + 1, line_, file_ );
    else
	fprintf( stderr, "%8d%8d 0x%.8x\n", count(), size(), this + 1 );
}


/*****************************************************************
 * mem_list_c::mem_list_c
 *
 * Construct a memory list object.  Only one of these should be
 * constructed for each application.  The constructor itself cannot
 * actually initialize any of its data members since other objects may
 * have called new/delete before this constructor is called.
 * Therefore, initialization of data members is done on demand in the
 * insert function.  The one thing we can do here is schedule the
 * reporting code using atexit().  Note that the report must be
 * generated _after_ all global static destructors have been invoked.
 */
mem_list_c::mem_list_c()
{
#ifndef GC
    if ( getenv( "OMIT_LEAKS" ) == NULL )
	atexit( mem_list_report );
#endif
}


/*****************************************************************
 * mem_list_c::~mem_list_c
 *
 * Destroy a memory list object.
 */
mem_list_c::~mem_list_c()
{
}


/*****************************************************************
 * mem_list_c::initialize
 *
 * Initialize the data members of the memory list object.  This cannot
 * be done in the constructor.  See the mem_list_c constructor for
 * more details.
 */
void
mem_list_c::initialize()
{
    head_ = NULL;

    char * cp = getenv( "WATCH_NEW" );
    if ( cp )
    {
	watch_new = (int)strtol( cp, NULL, 0 );
	fprintf( stderr, "Watching for new of %d\n", watch_new );
    }

    cp = getenv( "WATCH_DEL" );
    if ( cp )
    {
	watch_del = (int)strtol( cp, NULL, 0 );
	fprintf( stderr, "Watching for delete of %d\n", watch_del );
    }
}


/*****************************************************************
 * mem_list_c::insert
 *
 * Insert a new memory chunk into the linked list.  The memory chunk
 * must first be initialized, then linked into the list.  Note that a
 * constructor is not used to initialize the mem_elem_c because it is
 * not a self-contained object, it is a prefix to a memory unit.
 * Initialize user memory to a bogus, but identifiable, value.  Also
 * check if this allocation is interesting to the user and scream if
 * so.
 */
void
mem_list_c::insert( mem_elem_c * e, size_t sz, const char * f, int l )
{
    /* Initialize the mem_list_c without constructors. */
    if ( count == 0 )
	initialize();

    if ( ++count == watch_new )
	fprintf( stderr, "Memory element %d 0x%x just allocated %d bytes.\n",
		 watch_new, e + 1, sz );

    e->set_control( count, sz, 0, f, l );
    e->clear( 0xDA );		/* Data Allocated */
    e->set_cookie();

#ifndef GC
    e->prev_ = NULL;
    e->next_ = head();
    if ( head() )
	head()->prev_ = e;
    head_ = e;
#endif
}


/*****************************************************************
 * mem_list_c::remove
 *
 * Remove the given memory chunk from the linked list.  Note that the
 * application may deliver garbage to this routine inadvertently.  So
 * do our best to identify bogus memory chunks.  Also check that
 * memory has not been overrun.  Next, remove the chunk from the
 * list.  If it isn't in the list, scream.  Finally, zap memory to an
 * identifiable, but bogus, value (0xDF).
 */
void
mem_list_c::remove( mem_elem_c * e )
{
    assert( e->valid() );

    if ( e->count() == watch_del )
	cerr << __FILE__ << "(" << __LINE__ << "): "
	     << "Memory element " << watch_del << " " << e + 1
	     << " just deleted " << e->size() << " bytes." << endl;

#ifndef GC
    /* Remove the chunk from the list. */
    if ( e->prev_ )
	e->prev_->next_ = e->next_;
    else
	head_ = e->next_;

    if ( e->next_ )
	e->next_->prev_ = e->prev_;
#endif

    /* Set the freed bit. */
    e->set_free( 1 );
    e->clear( 0xDF );		/* Data Freed */
}


/*****************************************************************
 * mem_list_report
 *
 * Walk the list of allocated memory chunks printing a line for each.
 * Also note their identifier, size, and cookie status.
 */
void
mem_list_report()
{
    fprintf( stderr, "Memory Leak Report (count = %d, max size = %d)\n",
	     mem_list.count, mem_list.max_size );

    int cnt = 0;
    for ( mem_elem_c * ep = mem_list.head(); ep; ep = ep->next_, cnt++ )
	ep->report( cnt, 1 );

    fprintf( stderr, "Total memory items not deleted: %d.\n", cnt );
}


/*****************************************************************
 * mem_dbg
 */
extern "C" void
mem_dbg( mem_elem_c * m )
{
    if ( m != NULL && --m != NULL )
    {
	m->valid();
	m->report( 0, 0 );
    }
    else
	cerr << "Cannot print the NULL pointer." << endl;
}
